// Copyright 2016 by Sandro Santilli <strk@kbt.io>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA

package main

import (
	"crypto/sha1"
	"database/sql"
	"encoding/hex"
	"fmt"
	_ "github.com/lib/pq"
	"io"
	"log"
	"os"
	"path"
	"regexp"
	"strings"

	"github.com/spf13/pflag"
)

var DEFAULT_AUTHOR_ID int64
var DEFAULT_ASSIGNEE_ID int64

var tracdb, gogsdb *sql.DB
var gogsRepoId int64
var tracFilesPath string
var gogsRootPath string

func main() {

	LABELCOLOR_COMPONENT := "#fbca04"
	LABELCOLOR_PRIORITY := "#207de5"
	LABELCOLOR_SEVERITY := "#eb6420"
	LABELCOLOR_VERSION := "#009800"
	LABELCOLOR_TYPE := "#e11d21"
	LABELCOLOR_RESOLUTION := "#9e9e9e"

	defaultAssignee := pflag.String("default-assignee", "",
		"`username` to assign tickets to when trac assignee is not found in Gogs")
	defaultAuthor := pflag.String("default-author", "",
		"`username` to attribute content to when trac author is not found in Gogs")

	tracFilesPath_ := pflag.String("trac-files", "",
		"`path` to trac's files directory")
	gogsRootPath_ := pflag.String("gogs-root", "",
		"`path` to gogs's root directory")

	pflag.Usage = func() {
		fmt.Fprintf(os.Stderr,
			"Usage: %s [options] <trac_db> <gogs_db> <gogs_repo>\n",
			os.Args[0])
		fmt.Fprintf(os.Stderr, "Options:\n")
		pflag.PrintDefaults()
	}

	pflag.Parse()

	if pflag.NArg() < 3 {
		pflag.Usage()
		os.Exit(1)
	}

	var err error

	tracDbName := pflag.Arg(0)
	gogsDbName := pflag.Arg(1)
	gogsRepoName := pflag.Arg(2)

	// Check that paths are given both or none,
	// and that they are valid if given
	if *tracFilesPath_ != "" {
		if *gogsRootPath_ == "" {
			log.Fatal("Trac files and gogs data paths must be given both or none")
		}
		stat, err := os.Stat(*tracFilesPath_)
		if err != nil {
			log.Fatal(err)
		}
		if stat.IsDir() != true {
			log.Fatal("Trac files path is not a directory: ", *tracFilesPath_)
		}
		stat, err = os.Stat(*gogsRootPath_)
		if err != nil {
			log.Fatal(err)
		}
		if stat.IsDir() != true {
			log.Fatal("Gogs root path is not a directory: ", *gogsRootPath_)
		}
		tracFilesPath = *tracFilesPath_
		gogsRootPath = *gogsRootPath_
	} else if *gogsRootPath_ != "" {
		log.Fatal("Trac files and gogs data paths must be given both or none")
	} else {
		log.Println("WARNING: attachments will not be migrated, add --gogs-data and --trac-files for that")
	}

	tracdb, err = sql.Open("postgres", "dbname="+tracDbName)
	if err != nil {
		log.Fatal(err)
	}

	gogsdb, err = sql.Open("postgres", "dbname="+gogsDbName)
	if err != nil {
		log.Fatal(err)
	}

	gogsRepoId = findGogsRepoId(gogsdb, gogsRepoName)

	fmt.Printf("Importing trac db '%s' to repo '%s' (%d) in gogs db '%s'\n",
		tracDbName, gogsRepoName, gogsRepoId, gogsDbName)

	adminUser := findGogsAdminUser(gogsdb)
	adminUserId := findGogsUser(gogsdb, adminUser)

	DEFAULT_ASSIGNEE_ID = adminUserId
	if *defaultAssignee == "" {
		*defaultAssignee = adminUser
	} else {
		DEFAULT_ASSIGNEE_ID = findGogsUser(gogsdb, *defaultAssignee)
		if DEFAULT_ASSIGNEE_ID == -1 {
			log.Fatal("Cannot find gogs user ", *defaultAssignee)
		}
	}

	DEFAULT_AUTHOR_ID = adminUserId
	if *defaultAuthor == "" {
		*defaultAuthor = adminUser
	} else {
		DEFAULT_AUTHOR_ID = findGogsUser(gogsdb, *defaultAuthor)
		if DEFAULT_ASSIGNEE_ID == -1 {
			log.Fatal("Cannot find gogs user ", *defaultAuthor)
		}
	}

	fmt.Println("Default Gogs assignee is", *defaultAssignee, "(", DEFAULT_ASSIGNEE_ID, ")")
	fmt.Println("Default Gogs author is", *defaultAuthor, "(", DEFAULT_AUTHOR_ID, ")")

	// Import components as label "Component / <name>"

	rows, err := tracdb.Query("SELECT name FROM component")
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		lbl := "Component / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_COMPONENT)
	}

	// Import priorities as label "Priority / <name>"

	rows, err = tracdb.Query("SELECT distinct priority FROM ticket")
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		lbl := "Priority / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_PRIORITY)
	}

	// Import severities as label "Severity / <name>"

	rows, err = tracdb.Query("SELECT distinct severity FROM ticket")
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		lbl := "Severity / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_SEVERITY)
	}

	// Import versions as label "Version / <name>"

	rows, err = tracdb.Query(`SELECT distinct version FROM ticket UNION
SELECT name from version`)
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		if val == "" {
			continue
		}
		lbl := "Version / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_VERSION)
	}

	// Import types as label "Type / <name>"

	rows, err = tracdb.Query(`SELECT distinct type FROM ticket`)
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		if val == "" {
			continue
		}
		lbl := "Type / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_TYPE)
	}

	// Import resolutions as label "Resolution / <name>"

	rows, err = tracdb.Query(`
SELECT DISTINCT resolution FROM ticket WHERE trim(resolution) != '';
`)
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var val string
		if err := rows.Scan(&val); err != nil {
			log.Fatal(err)
		}
		if val == "" {
			continue
		}
		lbl := "Resolution / " + val
		fmt.Println("Adding label", lbl)
		addLabel(gogsdb, gogsRepoId, lbl, LABELCOLOR_RESOLUTION)
	}

	// Import milestones

	// NOTE: trac timestamps are to the microseconds, we just need seconds
	rows, err = tracdb.Query(`SELECT name,
int8(due*1e-6), int8(completed*1e-6), description FROM
milestone UNION select distinct(milestone),0,0,'' FROM ticket
WHERE milestone not in ( select name from milestone ) and milestone != ''`)
	if err != nil {
		log.Fatal(err)
	}
	for rows.Next() {
		var completed, due int64
		var nam, desc string
		if err := rows.Scan(&nam, &due, &completed, &desc); err != nil {
			log.Fatal(err)
		}
		fmt.Println("Adding milestone", nam)
		addMilestone(gogsdb, gogsRepoId, nam, desc, completed != 0, due, completed)
	}

	// Import tickets

	// NOTE: trac timestamps are to the microseconds, we just need seconds
	rows, err = tracdb.Query(`SELECT
t.id,t.type,
int8(t.time*1e-6),
t.component, t.severity, t.priority,
COALESCE(t.owner,''), t.reporter,
COALESCE(t.version,''),
COALESCE(t.milestone,''),
lower(COALESCE(t.status, '')),
COALESCE(t.resolution,''),
COALESCE(t.summary, ''),
COALESCE(t.description, '')
FROM ticket t order by id`)
	if err != nil {
		log.Fatal(err)
	}
	count := 0
	closed_count := 0
	for rows.Next() {
		var id, cre int64
		var comp, typ, sev, pri, own, rep, ver, mlst, stat, res, sum, desc string
		if err := rows.Scan(&id, &typ, &cre, &comp, &sev, &pri, &own, &rep,
			&ver, &mlst, &stat, &res, &sum, &desc); err != nil {
			log.Fatal(err)
		}
		fmt.Println("Adding ticket", id, " - ", sum)
		count++
		closed := stat == "closed"
		if closed {
			closed_count++
		}
		addTicket(gogsdb, gogsRepoId, id, typ, cre, comp, sev, pri, own,
			rep, ver, mlst, closed, res, sum, desc)
	}

	// Update issue count for repo
	if count > 0 {
		_, err = gogsdb.Exec(`UPDATE repository SET num_issues = num_issues+$1
	WHERE id = $2`, count, gogsRepoId)
		if err != nil {
			log.Fatal(err)
		}
	}
	if closed_count > 0 {
		_, err = gogsdb.Exec(`UPDATE repository
	SET num_closed_issues = num_closed_issues+$1
	WHERE id = $2`, closed_count, gogsRepoId)
		if err != nil {
			log.Fatal(err)
		}
	}

	// TODO: Update issue count for new labels

}

func tracTextToMarkdown(in string) string {

	out := in
	re := regexp.MustCompile("{{{([^\n]+?)}}}")
	out = re.ReplaceAllString(out, "`$1`")

	re = regexp.MustCompile("(?s){{{(.+?)}}}")
	out = re.ReplaceAllStringFunc(out, func(m string) string {
		lines := strings.Split(m, "\n")
		for i := range lines {
			l := lines[i]
			l = strings.Replace(l, "{{{", "", -1)
			l = strings.Replace(l, "}}}", "", -1)
			l = "\t" + l
			lines[i] = l
		}
		return strings.Join(lines, "\n")
	})

	// TODO:
	//    body.gsub!(/\=\=\=\=\s(.+?)\s\=\=\=\=/, '### \1')
	//    body.gsub!(/\=\=\=\s(.+?)\s\=\=\=/, '## \1')
	//    body.gsub!(/\=\=\s(.+?)\s\=\=/, '# \1')
	//    body.gsub!(/\=\s(.+?)\s\=[\s\n]*/, '')
	//    body.gsub!(/\[(http[^\s\[\]]+)\s([^\[\]]+)\]/, '[\2](\1)')
	//    body.gsub!(/\!(([A-Z][a-z0-9]+){2,})/, '\1')
	//    body.gsub!(/'''(.+)'''/, '*\1*')
	//    body.gsub!(/''(.+)''/, '_\1_')
	//    body.gsub!(/^\s\*/, '*')
	//    body.gsub!(/^\s\d\./, '1.')

	return out
}

// Return -1 if user is not found
func findGogsUser(db *sql.DB, name_or_address string) int64 {
	if strings.Trim(name_or_address, " ") == "" {
		return -1
	}
	sql := `SELECT id FROM public.user WHERE lower_name = $1 or email = $1`
	var id int64
	err := db.QueryRow(sql, name_or_address).Scan(&id)
	if err != nil {
		return -1
	}
	return id
}

func addIssueLabel(db *sql.DB, repoID int64, issueid int64, label string) {
	_, err := db.Exec(`INSERT INTO issue_label(issue_id, label_id)
  SELECT $1, (SELECT id FROM label where repo_id = $2 and name = $3)`,
		issueid, repoID, label)
	if err != nil {
		log.Fatal(err)
	}
}

func addTicketAttachment(gid, tid, tim, sz int64, author, fname, desc string) string {

	db := gogsdb

	comment := fmt.Sprintf("**Attachment** %s (%d bytes) added\n\n%s",
		fname, sz, desc)
	cid := addTicketComment(gid, tim, author, comment)

	tracpath := tracAttachmentPath(tid, fname)
	_, err := os.Stat(tracpath)
	if err != nil {
		log.Fatal(err)
	}

	elems := strings.Split(tracpath, "/")
	tracdir := elems[len(elems)-2]
	tracfil := elems[len(elems)-1]
	// 78ac is l33t for trac (horrible, I know)
	uuid := fmt.Sprintf("000078ac-%s-%s-%s-%s",
		tracdir[0:4], tracdir[4:8], tracdir[8:12],
		tracfil[0:12])
	gogspath := gogsAttachmentPath(uuid)
	// TODO: use a different uuid if file exists ?

	// TODO: avoid inserting record if uuid exist !
	_, err = db.Exec(`INSERT INTO attachment(
uuid, issue_id, comment_id, name, created_unix
) VALUES ($1, $2, $3, $4, $5)`, uuid, gid, cid, fname, tim)
	if err != nil {
		log.Fatal(err)
	}

	err = os.MkdirAll(path.Dir(gogspath), 0775)
	if err != nil {
		log.Fatal(err)
	}

	err = copyFile(tracpath, gogspath)
	if err != nil {
		log.Fatal(err)
	}

	return uuid
}

// comment is given in trac format
// returns gogs comment identifier
func addTicketComment(issue_id, time int64, author, comment string) int64 {

	var header []string
	db := gogsdb

	comment = tracTextToMarkdown(comment)

	// find users first, and tweak description to add missing users
	author_id := findGogsUser(db, author)
	if author_id == -1 {
		header = append(header, fmt.Sprintf("    Original comment by %s", author))
		author_id = DEFAULT_AUTHOR_ID
	}
	if len(header) > 0 {
		comment = fmt.Sprintf("%s\n\n%s", strings.Join(header, "\n"), comment)
	}

	row := db.QueryRow(`INSERT INTO comment(
type, issue_id, poster_id, content, created_unix, updated_unix)
  VALUES ( 0, $1, $2, $3, $4, $4 ) RETURNING id`,
		issue_id, author_id, comment, time)

	var cid int64
	err := row.Scan(&cid)
	if err != nil {
		log.Fatal(err)
	}

	return cid
}

// description is given in trac format
func addTicket(db *sql.DB, repoID int64,
	id int64, typ string, created int64,
	component string, severity string,
	priority string, owner string,
	reporter string, version string,
	milestone string, closed bool,
	resolution string, summary string,
	description string) {

	description = tracTextToMarkdown(description)

	var header []string

	// find users first, and tweak description to add missing users
	reporter_id := findGogsUser(db, reporter)
	if reporter_id == -1 {
		header = append(header, fmt.Sprintf("    Originally reported by %s", reporter))
		reporter_id = DEFAULT_AUTHOR_ID
	}
	var owner_id sql.NullString
	if owner != "" {
		tmp := findGogsUser(db, owner)
		if tmp == -1 {
			header = append(header, fmt.Sprintf("    Originally assigned to %s", owner))
			owner_id.String = fmt.Sprintf("%d", DEFAULT_ASSIGNEE_ID)
			owner_id.Valid = true
		} else {
			owner_id.String = fmt.Sprintf("%d", tmp)
			owner_id.Valid = true
		}
	} else {
		owner_id.Valid = false
	}
	if len(header) > 0 {
		description = fmt.Sprintf("%s\n\n%s", strings.Join(header, "\n"), description)
	}

	row := db.QueryRow(`INSERT INTO issue(index, repo_id, name, poster_id,
milestone_id, assignee_id, is_pull, is_closed, content, created_unix)
SELECT $1, $2, $3, $4,
(SELECT id FROM public.milestone WHERE repo_id = $2 AND name = $5),
$6,
false, $7, $8, $9 RETURNING id`,
		id, repoID, summary, reporter_id, milestone, owner_id,
		closed, description, created)

	var gid int64
	err := row.Scan(&gid)
	if err != nil {
		log.Fatal(err)
	}

	// Add labels

	var lbl string

	if component != "" {
		lbl = "Component / " + component
		addIssueLabel(db, repoID, gid, lbl)
	}

	if severity != "" {
		lbl = "Severity / " + severity
		addIssueLabel(db, repoID, gid, lbl)
	}

	if priority != "" {
		lbl = "Priority / " + priority
		addIssueLabel(db, repoID, gid, lbl)
	}

	if version != "" {
		lbl = "Version / " + version
		addIssueLabel(db, repoID, gid, lbl)
	}

	if resolution != "" {
		lbl = "Resolution / " + resolution
		addIssueLabel(db, repoID, gid, lbl)
	}

	if typ != "" {
		lbl = "Type / " + typ
		addIssueLabel(db, repoID, gid, lbl)
	}

	lastupdate := created

	// Add ticket attachments if requested

	if tracFilesPath != "" {

		rows, err := tracdb.Query(`
  SELECT
   int8(time*1e-6) tim,
   COALESCE(author, '') author,
   filename,
   description,
   size
  FROM attachment
  WHERE type = 'ticket'
  AND id = $1
  ORDER BY time asc
    `, id)
		if err != nil {
			log.Fatal(err)
		}

		for rows.Next() {
			var time, size int64
			var author, fname, desc string
			if err := rows.Scan(&time, &author, &fname, &desc, &size); err != nil {
				log.Fatal(err)
			}
			fmt.Println(" adding attachment by", author)
			if lastupdate > time {
				lastupdate = time
			}
			addTicketAttachment(gid, id, time, size, author, fname, desc)
		}

	}

	// Add ticket comments

	rows, err := tracdb.Query(`
SELECT
 int8(time*1e-6) tim,
 COALESCE(author, '') author,
 COALESCE(newvalue, '') newval
FROM ticket_change where ticket = $1
AND field = 'comment'
AND trim(COALESCE(newvalue, ''), ' ') != ''
ORDER BY time asc
  `, id)
	if err != nil {
		log.Fatal(err)
	}

	for rows.Next() {
		var time int64
		var author, comment string
		if err := rows.Scan(&time, &author, &comment); err != nil {
			log.Fatal(err)
		}
		fmt.Println(" adding comment by", author)
		if lastupdate > time {
			lastupdate = time
		}
		addTicketComment(gid, time, author, comment)
	}

	// Update issue modification time
	_, err = db.Exec(`UPDATE issue SET updated_unix = greatest(updated_unix,$1)
	WHERE id = $2`, lastupdate, gid)
	if err != nil {
		log.Fatal(err)
	}

}

func addMilestone(db *sql.DB, repoId int64, name string, content string, closed bool, deadline_ts int64, closed_ts int64) {
	_, err := db.Exec(`INSERT INTO
milestone(repo_id,name,content,is_closed,deadline_unix,closed_date_unix)
SELECT $1,$2::varchar,$3,$4,$5,$6 WHERE
NOT EXISTS ( SELECT * FROM milestone WHERE repo_id = $1 AND name = $2 )`,
		repoId, name, content, closed, deadline_ts, closed_ts)
	if err != nil {
		log.Fatal(err)
	}
}

func addLabel(db *sql.DB, repoId int64, label string, color string) {
	_, err := db.Exec(`INSERT INTO label(repo_id,name,color)
SELECT $1,$2::varchar, $3 WHERE
NOT EXISTS ( SELECT * FROM label WHERE repo_id = $1 AND name = $2 )`,
		repoId, label, color)
	if err != nil {
		log.Fatal(err)
	}
}

func findGogsAdminUser(db *sql.DB) string {
	sql := `
select lower_name from public.user where is_admin order by id limit 1;
`
	row := db.QueryRow(sql)

	var name string
	err := row.Scan(&name)
	if err != nil {
		fmt.Fprintf(os.Stderr, "No adming user found in Gogs\n")
		os.Exit(1)
	}

	return name
}

func findGogsRepoId(db *sql.DB, fullname string) int64 {

	comps := strings.Split(fullname, "/")
	if len(comps) < 2 {
		log.Fatal("Invalid gogs repo name (missing '/'): " + fullname)
	}
	owner := comps[0]
	repo := comps[1]

	sql := `
SELECT r.id FROM public.repository r, public.user u WHERE r.owner_id =
u.id AND u.name = $1 AND r.name = $2
`
	row := db.QueryRow(sql, owner, repo)

	var id int64
	err := row.Scan(&id)
	if err != nil {
		fmt.Fprintf(os.Stderr, "No "+fullname+" repository found in Gogs\n")
		os.Exit(1)
	}

	return id
}

func gogsAttachmentPath(uuid string) string {
	d1 := uuid[0:1]
	d2 := uuid[1:2]
	// TODO: seek for PATH under [attachment]
	//       in gogsRootPath/custom/conf/app.ini
	subpath := "data/attachments"
	return fmt.Sprintf("%s/%s/%s/%s/%s", gogsRootPath, subpath, d1, d2, uuid)
}

func tracAttachmentPath(tid int64, fname string) string {
	ticket_dir := encodeSha1(fmt.Sprintf("%d", tid))
	ticket_sub := ticket_dir[0:3]

	path_file := encodeSha1(fname)
	path_ext := path.Ext(fname)

	return fmt.Sprintf("%s/attachments/ticket/%s/%s/%s%s", tracFilesPath, ticket_sub, ticket_dir, path_file, path_ext)
}

// Encode string to sha1 hex value.
func encodeSha1(str string) string {
	h := sha1.New()
	h.Write([]byte(str))
	return hex.EncodeToString(h.Sum(nil))
}

func copyFile(src, dst string) error {
	in, err := os.Open(src)
	if err != nil {
		return err
	}
	defer in.Close()
	out, err := os.Create(dst)
	if err != nil {
		return err
	}
	defer out.Close()
	_, err = io.Copy(out, in)
	cerr := out.Close()
	if err != nil {
		return err
	}
	return cerr
}
